open Format

type info = Support.Error.info

type
    t = 
    Var of info * string
  | Con of info * int
  | SZero of info 
  | SOne of info 
  | Abs of info * string * Typ.aspect * Typ.t * t
  | App of info * t * t
  | Let of info * string * Typ.aspect * t * t
  | Case of info * Typ.t * t * t * t * t
  | Saferec of info * Typ.t 
  | Linrec of info * Typ.t 
  | Coerce of info * t * Typ.t

let gencase = function
 (fi,ty,e,(a,ea),(b,eb),(c,ec)) -> 
   if a < b & b < c 
      then Case(fi,ty,e,ea,eb,ec)
   else if a < c & c < b
      then Case(fi,ty,e,ea,ec,eb)
   else if b < c & c < a
      then Case(fi,ty,e,eb,ec,ea)
   else if b < a & a < c
      then Case(fi,ty,e,eb,ea,ec)
   else if c < b & b < a
      then Case(fi,ty,e,ec,eb,ea)
   else if c < a & a < b
      then Case(fi,ty,e,ec,ea,eb)
   else Support.Error.bug "gencase"      
 
type command = 
  Val of t
| ValBind of string * t
| AbbrevBind of info * string * Typ.t 
| ShowType of info * string
| Import of string
| Set of info * string * bool
| Do of info * string * string list

let expInfo = function
  Var(fi,_) -> fi | Con(fi,_) -> fi | SZero(fi) -> fi | SOne(fi) -> fi |
   Abs(fi,_,_,_,_) -> fi |  App(fi,_,_) -> fi |  
  Let(fi,_,_,_,_) -> fi | 
  Case(fi,_,_,_,_,_) -> fi | 
  Saferec(fi,_) -> fi | Linrec(fi,_) -> fi | 
  Coerce(fi,_,_) -> fi 

let elide_exps = 
  Support.Flags.createFlag "elide_exps"
    "Elide long subexpressions when printing expressions"
    true

let rec print = function
  Var(fi,x) -> 
    print_string x
| Con(fi,c) -> 
    print_int c
| SZero(fi) -> 
    print_string "SO"
| SOne(fi) -> 
    print_string "SI"
| Abs(fi,x,_,ty,e) -> 
    if !elide_exps then
      print_string "<fun>"
    else
      (print_string "(fun("; print_string x; print_string ":"; Typ.print ty;
       print_string ")"; print e; print_string ")")
| App(fi,e1,e2) -> 
    print_string "("; print e1; print_string " "; print e2;
    print_string ")"
| Let(fi,x,_,e,body) -> 
    if !elide_exps then
      print_string "<let exp>"
    else
      (print_string "(let "; print_string x; print_string "="; 
       print e; print_string " in "; print body; print_string ")")
| Coerce(fi,e1,t2) -> 
    print_string "("; print e1; print_string ":"; Typ.print t2;
    print_string ")"
| Case(fi,ty,e1,e2,e3,e4) -> 
    print_string "case[";Typ.print ty;print_string "] ";
    print e1; print_string " zero: ";print e2; 
    print_string " even: ";print e3;print_string " odd: ";print e4        
| Saferec(fi,ty) -> 
    print_string "saferec[";Typ.print ty;print_string "]"        
| Linrec(fi,ty) -> 
    print_string "linrec[";Typ.print ty;print_string "]"

(* ---------------------------------------------------------------------- *)
(* Substitution *)

let rec freevars = function
  Var(fi,x) -> 
    Support.StringSet.add x Support.StringSet.empty 
| Con(fi,c) -> 
    Support.StringSet.empty
| SZero(fi) ->
    Support.StringSet.empty
| SOne(fi) ->
    Support.StringSet.empty
| Abs(fi,x,_,ty,e1) -> 
    Support.StringSet.remove x (freevars e1)
| App(fi,e1,e2) -> 
    Support.StringSet.union (freevars e1) (freevars e2)
| Let(fi,x,_,e1,e2) -> 
    Support.StringSet.union
      (freevars e1)
      (Support.StringSet.remove x (freevars e2))
| Case(fi,ty,e1,e2,e3,e4) -> 
    Support.StringSet.union (freevars e1) 
  ( Support.StringSet.union (freevars e2) 
  ( Support.StringSet.union (freevars e3) (freevars e4)))
| Saferec(fi,ty) -> 
    Support.StringSet.empty
| Linrec(fi,ty) -> 
    Support.StringSet.empty
| Coerce(fi,e,t) ->
     freevars e

(* A substitution is a map from strings to elements of t *)
type substitution = t Support.StringMap.t

let emptysubst = Support.StringMap.empty

let extendsubst s x e = Support.StringMap.add x e s

(* If the substitution s contains a binding for the variable x, then return
   it; otherwise return x. *)
let applysubsttovar fi s x =
  try Support.StringMap.find x s
  with Not_found -> Var(fi,x)


(* Choose a fresh variable name as similar as possible to x but different 
   from the names in the domain of s or free in the range of s and also 
   different from the names in the set xs *)
let freshvar x s xs =
  let freevars_of_s = 
    Support.StringMap.fold
      (function x' -> function e -> function xs -> 
         Support.StringSet.add x' (Support.StringSet.union xs (freevars e)))
      s Support.StringSet.empty  
    in 
  let rec f x xs = if Support.StringSet.mem x xs then f (x ^ "'") xs else x
  in f x (Support.StringSet.union freevars_of_s xs)


(* ---------------------------- *)

let rec subst s = function
  Var(fi,x) -> 
     applysubsttovar fi s x
| Con(fi,c) -> 
    Con(fi,c)
| SZero(fi) -> 
    SZero(fi)
| SOne(fi) -> 
    SOne(fi)
| Abs(fi,x,a,ty,e1) -> 
    let x' = freshvar x s (freevars e1) in
    let s' = extendsubst s x (Var(Support.Error.unknown,x')) in
    Abs(fi, x',a, ty, subst s' e1)
| App(fi,e1,e2) -> 
    App(fi, subst s e1, subst s e2)
| Let(fi,x,a,e1,e2) -> 
    let x' = freshvar x s (freevars e2) in
    let s' = extendsubst s x (Var(Support.Error.unknown,x')) in
    Let(fi, x',a, subst s e1, subst s' e2)
| Case(fi,ty,e1,e2,e3,e4) -> 
    Case(fi,ty, subst s e1, subst s e2, subst s e3, subst s e4)
| Saferec(fi,ty) -> 
    Saferec(fi,ty)
| Linrec(fi,ty) -> 
    Linrec(fi,ty)
| Coerce(fi,e1,t2) -> 
    Coerce(fi, subst s e1, t2)

and subst1 x ex e = subst (extendsubst emptysubst x ex) e
and subst2 x1 ex1 x2 ex2 e =
  subst (extendsubst (extendsubst emptysubst x1 ex1) x2 ex2) e

let rename_binding ctx (x,e) =
  let x' = Typcontext.choose_fresh_var x ctx in
  if x = x' 
    then (x, e)
    else (x', subst1 x (Var(Support.Error.unknown,x')) e)

let rename_two_bindings ctx (x,y,e) =
  let x' = Typcontext.choose_fresh_var x ctx in
  let y' = Typcontext.choose_fresh_var y (Typcontext.dummybind x' ctx) in
  if x=x' & y=y'
    then (x, y, e)
    else (x', y', 
          subst2 x (Var(Support.Error.unknown,x'))
                 y (Var(Support.Error.unknown,y'))
            e)














