(*
 * Abstract syntax of expressions
 *)

type info = Support.Error.info

(* The datatype of expressions. The aspects in abstractions and let's
are inferred during typechecking and not visible to the user. They
might be later used for optimized evaluation *)

type
    t = 
    Var of info * string
  | Con of info * int
  | SZero of info
  | SOne of info
  | Abs of info * string * Typ.aspect * Typ.t * t
  | App of info * t * t
  | Let of info * string * Typ.aspect * t * t
  | Case of info * Typ.t * t * t * t * t
  | Saferec of info * Typ.t
  | Linrec of info * Typ.t 
  | Coerce of info * t * Typ.t

val gencase : info * Typ.t * t * (int*t) * (int*t) * (int*t) -> t

(* Return the "file info" of a given expression (for printing errors) *)
val expInfo: t -> info

(* Print an expression *)
val print : t -> unit

type substitution
val emptysubst : substitution
val extendsubst : substitution -> string -> t -> substitution

(* Substitute a list of variables with corresponding expressions 
   inside another expression *)
val subst : substitution -> t -> t

(* Substitute a variable with an expression inside another expression *)
val subst1 : string -> t -> t -> t
val subst2 : string -> t -> string -> t -> t -> t

(* Make sure that a bound variable name differs from any that appear in the
   context, renaming if necessary *)
val rename_binding : Typcontext.ctx -> (string * t) -> (string * t)

val rename_two_bindings : 
  Typcontext.ctx -> (string * string * t) -> (string * string * t)

(* Calculate the free variables of an expression *)
val freevars : t -> Support.StringSet.t


(* ------------------- *)

(* The datatype of top-level commands. (This is used only in the
   Toplevel module, but we define it here because the parser needs to
   build elements of this type.) *)

type command = 
  Val of t
| ValBind of string * t
| AbbrevBind of info * string * Typ.t 
| ShowType of info * string
| Import of string
| Set of info * string * bool
| Do of info * string * string list




