 (* 
 *  The lexical analyzer
 *)

{
open Support
open Error

type buildfun = Error.info -> Parser.token
let (symbolTable : (string,buildfun) Hashtbl.t) = Hashtbl.create 1024

let reservedWords = [
("->", fun i -> Parser.ARROW i);
("-o", fun i -> Parser.LINIMP i);
("$", fun i -> Parser.BOX i);
(";", fun i -> Parser.SEMI i);
(",", fun i -> Parser.COMMA i);
(":", fun i -> Parser.COLON i);
("=", fun i -> Parser.EQ i);
("[",fun i -> Parser.LSQUARE i);
("]",fun i -> Parser. RSQUARE i);
("(",fun i -> Parser.LPAREN i);
(")",fun i -> Parser. RPAREN i);
("fun", fun i -> Parser.FUN i);
("case", fun i -> Parser.CASE i);
("zero", fun i -> Parser.ZERO i);
("even", fun i -> Parser.EVEN i);
("odd", fun i -> Parser.ODD i);
("saferec", fun i -> Parser.SAFEREC i);
("linrec", fun i -> Parser.LINREC i);
("import", fun i -> Parser.IMPORT i);
("in", fun i -> Parser.IN i);
("N", fun i -> Parser.NAT i);
("SO", fun i -> Parser.SZERO i);
("SI", fun i -> Parser.SONE i);
("let", fun i -> Parser.LET i);
("NOW", fun i -> Parser.NOW i);
("RESET", fun i -> Parser.RESET i);
("SET", fun i -> Parser.SET i);
("showtype", fun i -> Parser.SHOWTYPE i);
("type", fun i -> Parser.TYPE i);
("val", fun i -> Parser.VAL i)
]
let _ =
  List.iter (fun (str,f) -> Hashtbl.add symbolTable str f) reservedWords

let createID i str =
  try (Hashtbl.find symbolTable str) i with _ -> Parser.ID {i=i;v=str}

(* Various counters *)

let lineno   = ref 1
and depth    = ref 0
and start    = ref 0

and filename = ref ""
and startLex = ref Error.unknown

let create inFile stream =
  if Filename.is_absolute inFile then filename := inFile
  else filename := Filename.concat (Sys.getcwd()) inFile;
  lineno := 1; start := 0; Lexing.from_channel stream

let newline lexbuf = incr lineno; start := (Lexing.lexeme_start lexbuf)

let info lexbuf =
  Error.create (!filename) (!lineno) (Lexing.lexeme_start lexbuf - !start)

let text = Lexing.lexeme

let stringBuffer = ref (String.create 2048)
let stringEnd = ref 0

let resetStr () = stringEnd := 0

let addStr ch =
  let x = !stringEnd in
  let buffer = !stringBuffer
in
  if x == String.length buffer then
    begin
      let newBuffer = String.create (x*2) in
      String.blit buffer 0 newBuffer 0 x;
      String.set newBuffer x ch;
      stringBuffer := newBuffer;
      stringEnd := x+1
    end
  else
    begin
      String.set buffer x ch;
      stringEnd := x+1
    end

let getStr () = String.sub (!stringBuffer) 0 (!stringEnd)

let extractLineno yytext offset =
  int_of_string (String.sub yytext offset (String.length yytext - offset))
}

rule main = parse
  [' ' '\009' '\012']+     { main lexbuf }

| [' ' '\009' '\012']*"\n" { newline lexbuf; main lexbuf }

| "-}" { Error.errAt (info lexbuf) "Unmatched end of comment" }

| "{-" { depth := 1; startLex := info lexbuf; comment lexbuf; main lexbuf }

| "# " ['0'-'9']+
    { lineno := extractLineno (text lexbuf) 2 - 1; getFile lexbuf }

| "# line " ['0'-'9']+
    { lineno := extractLineno (text lexbuf) 7 - 1; getFile lexbuf }

| ['0'-'9']+
    { Parser.NATV{i=info lexbuf; v=int_of_string (text lexbuf)} }


| ['A'-'Z' 'a'-'z' '_']
  ['A'-'Z' 'a'-'z' '_' '0'-'9' '\'']*
    { createID (info lexbuf) (text lexbuf) }

| ['~' '%' '\\' 'o' '+' '-' '<' '>' '=' '&' '|' ':' '@' '`']+
    { createID (info lexbuf) (text lexbuf) }

| ['*' '#' '/' '$' '!' '?' '^' '(' ')' '{' '}' '[' ']' '.' ';' '_' ',']
    { createID (info lexbuf) (text lexbuf) }

| "\"" { resetStr(); startLex := info lexbuf; string lexbuf }

| eof { Parser.EOF(info lexbuf) }

| _  { Error.errAt (info lexbuf) "Illegal character" }

and comment = parse
  "{-"
    { depth := succ !depth; comment lexbuf }
| "-}"
    { depth := pred !depth; if !depth > 0 then comment lexbuf }
| eof
    { Error.errAt (!startLex) "Comment not terminated" }
| [^ '\n']
    { comment lexbuf }
| "\n"
    { newline lexbuf; comment lexbuf }

and getFile = parse
  " "* "\"" { getName lexbuf }

and getName = parse
  [^ '"' '\n']+ { filename := (text lexbuf); finishName lexbuf }

and finishName = parse
  '"' [^ '\n']* { main lexbuf }

and string = parse
  '"'  { Parser.STRINGV {i = !startLex; v=getStr()} }
| '\\' { addStr(escaped lexbuf); string lexbuf }
| '\n' { addStr '\n'; newline lexbuf; string lexbuf }
| eof  { Error.errAt (!startLex) "String not terminated" }
| _    { addStr (Lexing.lexeme_char lexbuf 0); string lexbuf }

and escaped = parse
  'n'	 { '\n' }
| 't'	 { '\t' }
| '\\'	 { '\\' }
| '"'    { '\034'  }
| '\''	 { '\'' }
| ['0'-'9']['0'-'9']['0'-'9']
    {
      let x = int_of_string(text lexbuf) in
      if x > 255 then
	Error.errAt (info lexbuf) "Illegal character constant"
      else
	Char.chr x
    }
| [^ '"' '\\' 't' 'n' '\'']
    { Error.errAt (info lexbuf) "Illegal character constant" }

