(* 
   The implementation of the Support module actually defines three
   sub-modules: Error, Flags, and an internal module Misc used by
   Flags for internal data structures. 
*)

open Format

module Error = struct

exception Exit of int

type info = FI of string * int * int | UNKNOWN
type 'a withinfo = {i: info; v: 'a}

let unknown = UNKNOWN
let progName = FI(Array.get Sys.argv 0, 0, 0)
let create f l c = FI(f, l, c)

let errf f = 
  print_flush(); print_newline();
  open_vbox 0; 
  open_hvbox 0; f(); print_cut(); close_box(); print_newline();
  raise (Exit 1)

(* Special undocumented feature, so that error messages appearing in
   the documentation can be printed without hideous long automounter 
   pathnames *)
let fakeFileName = ref("")
let printErrorsAsIfInFile s = fakeFileName := s

let printInfo = function
  FI(f,l,c) ->
    print_string (if !fakeFileName = "" then f else !fakeFileName); 
    print_string ":"; 
    print_int l; print_string "."; 
    print_int c; print_string ":"
| UNKNOWN ->
    print_string "<Unknown file and line>: "

let errfAt fi f = errf(fun()-> printInfo fi; print_space(); f())

let err s = errf (fun()-> print_string s; print_newline())

let errAt fi s = errfAt fi (fun()-> print_string s; print_newline())

let bug s = err ("Compiler bug: " ^ s)

let notImplemented s = err ("Not implemented: " ^ s)

let bugf f = errf (fun()->
  print_string "Compiler bug: "; print_space(); f()
)

let bugfAt fi f = errfAt fi (fun()->
  print_string "Compiler bug: "; print_space(); f()
)

let bugAt fi s = errAt fi ("Compiler bug: " ^ s)

let warning s =
  print_string "Warning: "; print_string s;
  print_newline()

let warningAt fi s =
  printInfo fi; print_string " Warning: ";
  print_string s; print_newline()

end

(* ---------------------------------------------------------------------- *)

module Misc = struct

let iterateInt start finish step f res =
  if step = 0 then
    Error.bug "Misc.iterateInt"
  else if step > 0 then 
    let rec iterateUp index res =
      if index > finish then res
      else iterateUp (index + step) (f index res)
    in
      iterateUp start res
  else
    let rec iterateDown index res =
      if index < finish then res
      else iterateDown (index + step) (f index res)
    in
      iterateDown start res

module type S =
  sig
    type key
    type 'a t
    val empty: 'a t
    val add: key -> 'a -> 'a t -> 'a t
    val find: key -> 'a t -> 'a
    val remove: key -> 'a t -> 'a t
    val iter: (key -> 'a -> 'b) -> 'a t -> unit
    val fold: (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
    val union : 'a t -> 'a t -> 'a t
    val map : (key -> 'a -> 'b) -> 'a t -> 'b t
    val format :
      string -> string -> string -> (key -> 'a -> unit) -> 'a t -> unit
  end

module Make (Ord:Map.OrderedType) : (S with type key = Ord.t) =
  struct
    open Format
    module M = Map.Make(Ord)
    type key = Ord.t
    type 'a t = 'a M.t
    let empty = M.empty
    let add = M.add
    let find = M.find
    let remove = M.remove
    let iter = M.iter
    let fold = M.fold
    let union m1 m2 = M.fold M.add m2 m1
    let map f m = M.fold (fun k e m -> M.add k (f k e) m) m M.empty

    let format start finish sep p m =
      let rec pr k e needSep =
	if needSep then
	  (print_string sep; print_cut(); p k e; true)
	else
	  (p k e; true)
    in
      print_string start;
      open_hovbox 0; M.fold pr m false;
      print_string finish;
      close_box()
  end

module StringMap =
  Make(struct
    type t = string
    let compare = compare
  end)

module IntMap =
  Make(struct
    type t = int
    let compare = (-)
  end)

let ist start finish sep p l =
  let rec pr needSep = function
    [] -> ()
  | (h :: t) ->
    if needSep then
      (print_string sep;
       print_cut(); p h;
       pr true t)
    else
      (p h; pr true t)
in
  print_string start;
  open_hovbox 0;
  pr false l;
  print_string finish;
  close_box()

let printList start finish sep p os l =
  let rec pr needSep = function
    [] -> ()
  | (h :: t) ->
    if needSep then
      (output_string os sep; p os h; pr true t)
    else
      (p os h; pr true t)
in
  output_string os start; pr false l; output_string os finish

let itFold f b l =
  let rec iterate i b = function
    [] -> b
  | (x::xs) -> iterate (succ i) (f i x b) xs
in
  iterate 0 b l

end

module StringMap = Misc.StringMap

module StringSet =
  Set.Make(struct
    type t = string
    let compare = compare
  end)

(* ---------------------------------------------------------------------- *)

module Flags = struct

(*
 * Flags, accessible dynamically by name.
 *)

(* In this module, we use the lower-level Prim error reporting instead
   of the functions in main.mli to avoid circular linkage dependencies *)

(* (Current value, documentation, default value, currently-enabled *)
type flag = bool ref * string * bool * bool

let flags = ref (Misc.StringMap.empty : flag Misc.StringMap.t);;

let createFlagInternal name doc init switch =
  let f = ref init
in
  try
    Misc.StringMap.find name (!flags);
    Error.err("Multiple definition of flag: " ^ name)
  with Not_found ->
    (flags := Misc.StringMap.add name (f,doc,init,switch) (!flags); 
     f)

let createFlagIfEnabled name doc init switch =
  createFlagInternal name doc init switch

let createFlag name doc init =
  createFlagInternal name doc init true

let setFlag fi b name = 
  try
    let (f,_,_,enabled) = Misc.StringMap.find name (!flags) in
      if enabled then f := b
      else Error.warning ("Flag " ^ name ^ " is disabled")
  with Not_found ->
    Error.errAt fi ("Unknown flag: " ^ name)

let expert = createFlag "expert" 
  "Enable some compiler debugging features" false

let printFlag name (_, doc, init, enabled) =
  if String.get doc 0 <> '*' or !expert then begin
    print_string "  ";
    print_string name;
    if not enabled then print_string " (disabled)";
    print_string "\n    ";
    print_string doc;
    print_string "\n    default:";
    if init then print_string "true" else print_string "false";
    print_newline()
  end

let printFlags () =
  Misc.StringMap.format "Internal flags:\n" "" "" printFlag (!flags)

(* ---------------------------------------------------------------------- *)

type cmd = string * (Error.info -> string list -> unit)

let cmds = ref (Misc.StringMap.empty : cmd Misc.StringMap.t);;

let createCmd name doc beh =
  try
    Misc.StringMap.find name (!cmds);
    Error.err("Multiple definition of command: " ^ name)
  with Not_found ->
    cmds := Misc.StringMap.add name (doc,beh) (!cmds)

let doCmd fi name args = 
  try
    let (_,beh) = Misc.StringMap.find name (!cmds) in beh fi args
  with Not_found ->
    Error.errAt fi ("Unknown command: " ^ name)

(* Install commands for setting and resetting flags, and for printing
   out the documentation of all flags and commands *)

let () = createCmd "set" "Set one or more flags to TRUE"
  (fun fi l -> List.iter (fun f -> setFlag fi true f) l)

let () = createCmd "reset" "Set one or more flags to FALSE"
  (fun fi l -> List.iter (fun f -> setFlag fi false f) l)

let printCmd name (doc, beh) =
  if String.get doc 0 <> '*' or !expert then begin
    print_string "  ";
    print_string name;
    print_string "\n    ";
    print_string doc;
    print_newline()
  end

let printCmds () =
  Misc.StringMap.format "Internal commands:\n" "" "" printCmd (!cmds)

(* ---------------------------------------------------------------------- *)

let () = createCmd "help" "Print this usage message"
  (fun _ _ -> printCmds(); printFlags())

(* ---------------------------------------------------------------------- *)

let () = createCmd "printErrorsAsIfInFile" 
  "*Print errors as if current source were named FILE (default 'example.f')"
  (fun _ ->
     function [s] -> Error.printErrorsAsIfInFile s
     | _ -> Error.printErrorsAsIfInFile "example.f")

end

let rec assoc3 v = function
    [] -> raise Not_found
  | (v',x,y)::rest -> if v=v' then (x,y) else assoc3 v rest

let rec mem_assoc3 v = function
    [] -> false
  | (v',x,y)::rest -> v=v' or mem_assoc3 v rest

